from typing import Any, Callable, TypeAliasType

from more_itertools import flatten

from testing import Exercise, points_to_deduct

def has_annotation_callable(f: Callable, param_types: list[type], return_type: object = None) -> bool:
    a = f.__annotations__
    ret = a['return'] if 'return' in a else None
    if ret != return_type:
        return False

    params = list(a.values())[:-1] if 'return' in a else list(a.values())
    for param, param_type in zip(params, param_types):
        while isinstance(param, TypeAliasType):
            param = param.__value__
        while isinstance(param_type, TypeAliasType):
            param_type = param_type.__value__
        if param != param_type:
            return False
    return True

def format_type(t) -> str:
    while isinstance(t, TypeAliasType):
        t = t.__value__
    match t:
        case type():
            return t.__name__.split('.')[-1]
        case _:
            return str(t).split('.')[-1]
        
def format_callable(f: Callable[..., Any]) -> str:
    a = list(f.__annotations__.values() if 'return' in f.__annotations__ else f.__annotations__.values() + [None])
    a = [format_type(t) for t in a]
    return f"{f.__name__}({", ".join(a[:-1])}) -> {a[-1]}"


def assert_annotation_callable(f: Callable[..., Any], param_types: list[type], return_type: object = None):
    assert has_annotation_callable(f, param_types, return_type), f"{f.__name__}({", ".join(format_type(p) for p in param_types)}) -> {format_type(return_type)} != {format_callable(f)}"
    
def format(ob: Exercise | object | Callable[..., Any]) -> list[str]:
    match ob:
        case Exercise():
            out = [f"# {ob.id} ({ob.points:.2g} / {ob.max_points:.2g})"]
            return out + list(flatten(map(format, ob.tasks)))
        case object() if hasattr(ob, 'is_task'):
            return format_task("##", ob)
        case Callable() if hasattr(ob, 'is_test'):
            return format_test(ob)
        case _:
            return []
        
def format_test(test: Callable[..., Any]) -> list[str]:
    if not test.has_failed:
        return []
    return [f"- {test.msg}  [`-{test.to_deduct():.2g}p`]"]
    
def format_task(prefix: str, task: object) -> list[str]:
    out = [f"{prefix} {task.header} - ({(task.max_points - points_to_deduct(task)):.2g} / {task.max_points:.2g})"]
    for t in task.tasks:
        out += format_task(prefix + "#", t)
    for t in task.tests:
        out += format_test(t)
    return out